local tablet = AceLibrary("Tablet-2.0")
local deformat = AceLibrary("Deformat-2.0")
local babbleclass = AceLibrary("Babble-Class-2.2")
local crayon = AceLibrary("Crayon-2.0")
local tourist = AceLibrary("Tourist-2.0")
local quixote = AceLibrary("Quixote-1.0")

local L = AceLibrary("AceLocale-2.2"):new("QuestsFu")

QuestsFu = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceHook-2.1", "AceConsole-2.0", "AceDB-2.0", "FuBarPlugin-2.0")

QuestsFu:RegisterDB("QuestsFuDB")
QuestsFu:RegisterDefaults('profile', {
	showLevelsGame = true,
	showLevelsTablet = true,
	showLevelsZone = true,
	showDifficulty = true,
	showImpossible = true,
	showArea = true,
	showCurrentAreaOnly = false,
	showClassAnyway = true,
	showCurrentAreaDescriptionOnly = false,
	showDescription = true,
	showCompletedObjectives = false,
	showInItemTooltip = true,
	colorObjectives = true,
	wrapQuests = true,
	showTextOpt = {
		total = true,
		current = true,
		complete = false,
		lastmessage = true,
	},
	useOwnTracker = true,
	tracker_data = {
		fontSizePercent=0.7, detached=true, transparency=0, 
	},
})
QuestsFu:RegisterDefaults('char', {
	hidden = {},
	watchedQuests = {},
})

QuestsFu.version = "2.0." .. string.sub("$Revision: 19234 $", 12, -3)
QuestsFu.date = string.sub("$Date: 2006-12-05 16:31:18 -0500 (Tue, 05 Dec 2006) $", 8, 17)
QuestsFu.hasIcon = true
QuestsFu.clickableTooltip = true
QuestsFu.cannotHideText = true

local optionsTable = {
	handler = QuestsFu,
	type = 'group',
	args = {
		show = {
			type = 'group', 
			name = L["OPT_show"],
			desc = L["OPT_show_d"],
			args = {
				text = {
					type = 'group', 
					name = L["OPT_text"],
					desc = L["OPT_text_d"],
					args = {
						current = {
							type = 'toggle', 
							name = L["OPT_current"],
							desc = L["OPT_current_d"],
							get = "IsShowingTextCurrent",
							set = "ToggleShowingTextCurrent",
						},
						complete = {
							type = 'toggle', 
							name = L["OPT_complete"],
							desc = L["OPT_complete_d"],
							get = "IsShowingTextComplete",
							set = "ToggleShowingTextComplete",
						},
						total = {
							type = 'toggle', 
							name = L["OPT_total"],
							desc = L["OPT_total_d"],
							get = "IsShowingTextTotal",
							set = "ToggleShowingTextTotal",
						},
						lastmessage = {
							type = 'toggle', 
							name = L["OPT_lastmessage"],
							desc = L["OPT_lastmessage_d"],
							get = "IsShowingTextLastMessage",
							set = "ToggleShowingTextLastMessage",
						},
					}
				},
				levels = {
					type = 'group',
					name = L["OPT_levels"],
					desc = L["OPT_levels_d"],
					args = {
						game = {
							type = "toggle", name = L["OPT_levels_game"],
							desc = L["OPT_levels_game_d"],
							get = "IsShowingLevelsGame",
							set = "ToggleShowingLevelsGame",
						},
						tablet = {
							type = "toggle", name = L["OPT_levels_tablet"],
							desc = L["OPT_levels_tablet_d"],
							get = "IsShowingLevelsTablet",
							set = "ToggleShowingLevelsTablet",
						},
						zone = {
							type = "toggle", name = L["OPT_levels_zone"],
							desc = L["OPT_levels_zone_d"],
							get = "IsShowingLevelsZone",
							set = "ToggleShowingLevelsZone",
						},
					},
				},
				diff = {
					type = 'toggle', 
					name = L["OPT_diff"],
					desc = L["OPT_diff_d"],
					get = "IsShowingDifficulty",
					set = "ToggleShowingDifficulty",
				},
				impossible = {
					type = 'toggle', 
					name = L["OPT_impossible"],
					desc = L["OPT_impossible_d"],
					get = "IsShowingImpossible",
					set = "ToggleShowingImpossible",
				},
				area = {
					type = 'toggle', 
					name = L["OPT_area"],
					desc = L["OPT_area_d"],
					get = "IsShowingArea",
					set = "ToggleShowingArea",
				},
				caonly = {
					type = 'toggle', 
					name = L["OPT_caonly"],
					desc = L["OPT_caonly_d"],
					get = "IsShowingCurrentAreaOnly",
					set = "ToggleShowingCurrentAreaOnly",
				},
				classanyway = {
					type = 'toggle', 
					name = L["OPT_classanyway"],
					desc = L["OPT_classanyway_d"],
					get = "IsShowingClassAnyway",
					set = "ToggleShowingClassAnyway",
				},
				description = {
					type = 'toggle', 
					name = L["OPT_description"],
					desc = L["OPT_description_d"],
					get = "IsShowingDescription",
					set = "ToggleShowingDescription",
				},
				cadonly = {
					type = 'toggle', 
					name = L["OPT_cadonly"],
					desc = L["OPT_cadonly_d"],
					get = "IsShowingCurrentAreaDescriptionOnly",
					set = "ToggleShowingCurrentAreaDescriptionOnly",
				},
				completed= {
					type = 'toggle', 
					name = L["OPT_completed"],
					desc = L["OPT_completed_d"],
					get = "IsShowingCompletedObjectives",
					set = "ToggleShowingCompletedObjectives",
				},
				tooltip = {
					type = 'toggle', 
					name = L["OPT_tooltip"],
					desc = L["OPT_tooltip_d"],
					get = "IsShowingInItemTooltip",
					set = "ToggleShowingInItemTooltip",
				},
				colorobj = {
					type = 'toggle', 
					name = L["OPT_colorobj"],
					desc = L["OPT_colorobj_d"],
					get = "IsColoringObjectives",
					set = "ToggleColoringObjectives",
				},
				wrap = {
					type = 'toggle', 
					name = L["OPT_wrap"],
					desc = L["OPT_wrap_d"],
					get = "IsWrappingQuests",
					set = "ToggleWrappingQuests",
				},
				tracker = {
					type = 'toggle', name = L["OPT_tracker"],
					desc = L["OPT_tracker_d"],
					get = function() return QuestsFu.db.profile.useOwnTracker end,
					set = function(t)
						QuestsFu.db.profile.useOwnTracker = t
						if QuestsFu.db.profile.useOwnTracker then
							if tablet:IsRegistered('QuestsFu_Tracker') then tablet:Open('QuestsFu_Tracker') end
							QuestsFu:UpdateTracker()
						else tablet:Close('QuestsFu_Tracker'); QuestWatchFrame:Show() end
					end,
				}
			}
		}
	}
}

QuestsFu:RegisterChatCommand({"/questsfu", "/fuq" }, optionsTable)
QuestsFu.OnMenuRequest = optionsTable

function QuestsFu:ToggleShowingTextComplete()
	self.db.profile.showTextOpt.complete = not self.db.profile.showTextOpt.complete
	self:UpdateText()
	return self.db.profile.showTextOpt.complete
end
function QuestsFu:ToggleShowingTextCurrent()
	self.db.profile.showTextOpt.current = not self.db.profile.showTextOpt.current
	self:UpdateText()
	return self.db.profile.showTextOpt.current
end
function QuestsFu:ToggleShowingTextTotal()
	self.db.profile.showTextOpt.total = not self.db.profile.showTextOpt.total
	self:UpdateText()
	return self.db.profile.showTextOpt.total
end
function QuestsFu:ToggleShowingTextLastMessage()
	self.db.profile.showTextOpt.lastmessage = not self.db.profile.showTextOpt.lastmessage
	self:UpdateText()
	return self.db.profile.showTextOpt.lastmessage
end

function QuestsFu:IsShowingTextComplete()
	return self.db.profile.showTextOpt.complete
end
function QuestsFu:IsShowingTextCurrent()
	return self.db.profile.showTextOpt.current
end
function QuestsFu:IsShowingTextTotal()
	return self.db.profile.showTextOpt.total
end
function QuestsFu:IsShowingTextLastMessage()
	return self.db.profile.showTextOpt.lastmessage
end

function QuestsFu:IsShowingLevelsGame()
	return self.db.profile.showLevelsGame
end

function QuestsFu:IsShowingLevelsTablet()
	return self.db.profile.showLevelsTablet
end

function QuestsFu:IsShowingLevelsZone()
	return self.db.profile.showLevelsZone
end

function QuestsFu:ToggleShowingLevelsGame(silent)
	self.db.profile.showLevelsGame = not self.db.profile.showLevelsGame
	if self.db.profile.showLevelsGame then
		self:RegisterEvent("GOSSIP_SHOW", "OnGossipShow")
		self:RegisterEvent("QUEST_GREETING", "OnQuestGreeting")
		self:SecureHook("QuestLog_Update")
	else
		self:UnregisterEvent("GOSSIP_SHOW", "OnGossipShow")
		self:UnregisterEvent("QUEST_GREETING", "OnQuestGreeting")
		self:Unhook("QuestLog_Update")
	end
	self:UpdateData()
	return self.db.profile.showLevelsGame
end

function QuestsFu:ToggleShowingLevelsTablet(silent)
	self.db.profile.showLevelsTablet = not self.db.profile.showLevelsTablet
	self:UpdateTooltip()
	return self.db.profile.showLevelsTablet
end

function QuestsFu:ToggleShowingLevelsZone(silent)
	self.db.profile.showLevelsZone = not self.db.profile.showLevelsZone
	self:UpdateTooltip()
	return self.db.profile.showLevelsZone
end

function QuestsFu:IsShowingDifficulty()
	return self.db.profile.showDifficulty
end

function QuestsFu:ToggleShowingDifficulty(silent)
	self.db.profile.showDifficulty = not self.db.profile.showDifficulty
	self:UpdateTooltip()
	return self.db.profile.showDifficulty
end

function QuestsFu:IsShowingImpossible()
	return self.db.profile.showImpossible
end

function QuestsFu:ToggleShowingImpossible(silent)
	self.db.profile.showImpossible = not self.db.profile.showImpossible
	self:UpdateTooltip()
	return self.db.profile.showImpossible
end

function QuestsFu:IsShowingArea()
	return self.db.profile.showArea
end

function QuestsFu:ToggleShowingArea(silent)
	self.db.profile.showArea = not self.db.profile.showArea
	self:UpdateTooltip()
	return self.db.profile.showArea
end

function QuestsFu:IsShowingCurrentAreaOnly()
	return self.db.profile.showCurrentAreaOnly
end

function QuestsFu:ToggleShowingCurrentAreaOnly(silent)
	self.db.profile.showCurrentAreaOnly = not self.db.profile.showCurrentAreaOnly
	self:UpdateTooltip()
	return self.db.profile.showCurrentAreaOnly
end

function QuestsFu:IsShowingClassAnyway()
	return self.db.profile.showClassAnyway
end

function QuestsFu:ToggleShowingClassAnyway(silent)
	self.db.profile.showClassAnyway = not self.db.profile.showClassAnyway
	self:UpdateTooltip()
	return self.db.profile.showClassAnyway
end

function QuestsFu:IsShowingCurrentAreaDescriptionOnly()
	return self.db.profile.showCurrentAreaDescriptionOnly
end

function QuestsFu:ToggleShowingCurrentAreaDescriptionOnly(silent)
	self.db.profile.showCurrentAreaDescriptionOnly = not self.db.profile.showCurrentAreaDescriptionOnly
	self:UpdateTooltip()
	return self.db.profile.showCurrentAreaDescriptionOnly
end

function QuestsFu:IsShowingDescription()
	return self.db.profile.showDescription
end

function QuestsFu:ToggleShowingDescription(silent)
	self.db.profile.showDescription = not self.db.profile.showDescription
	self:UpdateTooltip()
	return self.db.profile.showDescription
end

function QuestsFu:IsColoringObjectives()
	return self.db.profile.colorObjectives
end

function QuestsFu:ToggleColoringObjectives(silent)
	self.db.profile.colorObjectives = not self.db.profile.colorObjectives
	self:UpdateTooltip()
	return self.db.profile.colorObjectives
end

function QuestsFu:IsShowingCompletedObjectives()
	return self.db.profile.showCompletedObjectives
end

function QuestsFu:ToggleShowingCompletedObjectives(silent)
	self.db.profile.showCompletedObjectives = not self.db.profile.showCompletedObjectives
	self:UpdateTooltip()
	return self.db.profile.showCompletedObjectives
end

function QuestsFu:IsShowingInItemTooltip()
	return self.db.profile.showInItemTooltip
end

function QuestsFu:ToggleShowingInItemTooltip(silent)
	self.db.profile.showInItemTooltip = not self.db.profile.showInItemTooltip
	self:UpdateTooltip()
	return self.db.profile.showInItemTooltip
end

function QuestsFu:IsWrappingQuests()
	return self.db.profile.wrapQuests
end

function QuestsFu:ToggleWrappingQuests(silent)
	self.db.profile.wrapQuests = not self.db.profile.wrapQuests
	self:UpdateTooltip()
	return self.db.profile.wrapQuests
end

function QuestsFu:OnInitialize()
	if type(self.db.profile.showText) == 'table' then
		--FuBarPlugin-2.0 appropriated .showText.  Switch it out.
		self.db.profile.showTextOps = self.db.profile.showText
		self.db.profile.showText = nil
	end
	
	self.lastuimessage = ""
	self.lastquestmessage = ""
	
	self.loadedWatchedQuests = false
end

function QuestsFu:OnEnable()
	self:RegisterEvent("Quixote_Update", "Update")
	self:RegisterEvent("PLAYER_LEVEL_UP", "UpdateTooltip") -- Quest difficulty colors can change on level; just redraw the tooltip in case they have it open while levelling.
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "UpdateTooltip")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT", "OnMouseOverUnit")
	self:RegisterEvent("UI_INFO_MESSAGE","OnUIInfoMessage")
	self:SecureHook("AddQuestWatch")
	self:SecureHook("RemoveQuestWatch")
	self:RegisterEvent("Quixote_Quest_Lost", "Quest_Lost")

	self:SecureHook("QuestWatch_Update", "UpdateTracker")
    self:Hook("GetLFGTypeEntries", true)

	self:HookTooltips()

	self:ScheduleEvent(self.LoadWatchedQuests, 1, self)

	if self:IsShowingLevelsGame() then
		self:RegisterEvent("GOSSIP_SHOW", "OnGossipShow")
		self:RegisterEvent("QUEST_GREETING", "OnQuestGreeting")
		self:SecureHook("QuestLog_Update")
	end
end

function QuestsFu:QuestLog_Update()
	for i=1, QUESTS_DISPLAYED, 1 do
		questIndex = i + FauxScrollFrame_GetOffset(QuestLogListScrollFrame)
		questLogTitle = getglobal("QuestLogTitle"..i)
		title, level, tag = quixote:GetQuest(questIndex)
		if title then
			questLogTitle:SetText(self:MakeTag(level, tag)..title)
		end
	end
end

function QuestsFu:GetLFGTypeEntries(index)
	if index and tonumber(index) then
		index = tonumber(index)
		local original = {self.hooks["GetLFGTypeEntries"](index)}
		for i, v in ipairs(original) do
			if index == 2 or index == 3 or index == 5 then
				if (index == 5 and tourist:IsZone(v)) or tourist:IsInstance(v) then
					local low, high = tourist:GetLevel(v)
					if low and high then
						original[i] = v .. " ("..tostring(low).."-"..tostring(high)..")"
					end
				end
			elseif index == 4 then
				local _, level = quixote:GetQuestByName(v)
				if level and tonumber(level) then
					original[i] = "["..tostring(level).."] "..v
				end
			end
		end
		return unpack(original)
	end
	return self.hooks["GetLFGTypeEntries"](index)
end

function QuestsFu:MakeTag(level, tag)
	if tag then
		if tag == '+' then tag = L["TAG_ELITE"]
		elseif tag == 'd' then tag = L["TAG_DUNGEON"]
		elseif tag == 'p' then tag = L["TAG_PVP"]
		elseif tag == 'r' then tag = L["TAG_RAID"]
		end
	else
		tag = ''
	end
	return "[" .. level .. tag .. "] "
end

--This is pretty much how Minimalist does it, which it credits to AutoSelect.
function QuestsFu:OnGossipShow()
	local buttonindex = 1
	local list, button
	if (GetGossipAvailableQuests()) then
		list = {GetGossipAvailableQuests()}
		for i = 2,getn(list),2 do
			button = getglobal("GossipTitleButton"..(buttonindex))
			button:SetText(format('[%d] %s',list[i],list[i-1]))
			buttonindex = buttonindex + 1
		end
		buttonindex = buttonindex + 1
		list = nil
	end
	if (GetGossipActiveQuests()) then
		list = {GetGossipActiveQuests()}
		for i = 2,getn(list),2 do
			button = getglobal("GossipTitleButton"..(buttonindex))
			button:SetText(format('[%d] %s',list[i],list[i-1]))
			buttonindex = buttonindex + 1
		end
		list = nil
	end
end

--This is pretty much how Minimalist does it, which it credits to AutoSelect.
function QuestsFu:OnQuestGreeting()
	local numactive,numavailable = GetNumActiveQuests(), GetNumAvailableQuests()
	local title, level, button
	local o,GetTitle,GetLevel = 0,GetActiveTitle,GetActiveLevel
	for i=1, numactive + numavailable do
		if i == numactive + 1 then
			o,GetTitle,GetLevel = numactive,GetAvailableTitle,GetAvailableLevel
		end
		title,level = GetTitle(i-o), GetLevel(i-o)
		button = getglobal("QuestTitleButton"..i)
		button:SetText(format('[%d] %s',level,title))
	end
end

function QuestsFu:OnUIInfoMessage()
	self.lastuimessage = arg1
	self:ScheduleEvent(self.ClearUIInfoMessage, 0.5, self)
end

function QuestsFu:ClearUIInfoMessage()
	self.lastuimessage = ""
end

function QuestsFu:OnMouseOverUnit()
	-- Make sure that this is wanted, and that the tooltip is actually read/writeable.
	if self:IsShowingInItemTooltip() and self:CanMessWithGameTooltip() then
		--Comes with color codes.  Strip 'em out.
		--local thisMob = string.gsub(getglobal('GameTooltipTextLeft1'):GetText(),"|c........(.*)|?r?","%1")
		local q,i,n = quixote:IsQuestMob(UnitName('mouseover'))
		if q then
			GameTooltip:AddDoubleLine(q, i..'/'..n, 1, 1, 1, crayon:GetThresholdColor(i/n))
			GameTooltip:Show()
		end
	end
end

local getLinkFunctions = {
	SetLootItem			= GetLootSlotLink,
	SetLootRollItem		= function(id) return GetLootRollItemLink(id) end,
	SetHyperlink		= function(link) return link end,
	SetAuctionItem		= GetAuctionItemLink,
	SetBagItem			= GetContainerItemLink,
	SetCraftItem		= function(skill, id) return id and GetCraftReagentItemLink(skill, id) or GetCraftItemLink(skill) end,
	SetInventoryItem	= function(type, slot) if not this then return end return type and GetInventoryItemLink(type, slot) or GetContainerItemLink(BANK_CONTAINER,this:GetID()) end,
	SetQuestItem		= GetQuestItemLink,
	SetQuestLogItem		= GetQuestLogItemLink,
	SetTradePlayerItem	= GetTradePlayerItemLink,
	SetTradeSkillItem	= function(skill, id) return (id) and GetTradeSkillReagentItemLink(skill, id) or GetTradeSkillItemLink(skill) end,
	SetTradeTargetItem	= GetTradeTargetItemLink,
}

function QuestsFu:LinkToName(link)
	if link then
		local _, _, id = string.find(link, "(%d+):")
		return id and GetItemInfo(id) or nil
	end
end

function QuestsFu:HookTooltips()
	for o, l in pairs(getLinkFunctions) do
		local originalFunction, linkFunction = o, l  -- Get these into a local context, to avoid errors.  This drove me nuts.
		self:SecureHook(GameTooltip, originalFunction, function(tooltip, arg1, arg2, arg3)
			local q,i,n = quixote:IsQuestItem(self:LinkToName(linkFunction(arg1, arg2, arg3)))
			if q then
				GameTooltip:AddDoubleLine(q, i..'/'..n, 1, 1, 1, crayon:GetThresholdColor(i/n))
				GameTooltip:Show()
			end
		end)
	end
end

function QuestsFu:CanMessWithGameTooltip()
	if GameTooltip and getglobal('GameTooltipTextLeft1'):IsVisible() and getglobal('GameTooltipTextLeft1'):GetText() ~= nil then
		return true
	else
		return false
	end
end

function QuestsFu:OnDataUpdate()
	self.lastquestmessage = self.lastuimessage
	self:SaveWatchedQuests()
	--A watched quest may have gone away because of the update.  (The game doesn't call RemoveQuestWatch for it...)
end

function QuestsFu:OnTextUpdate()
	local text = ""
	local maxQuests = MAX_QUESTS
	local r,g,b = GameFontNormal:GetTextColor()
	local color = string.format("%02X%02X%02X", r * 255, g * 255, b * 255)
	local numQuests, numQuestsDone = quixote:GetNumQuests()
	if self:IsShowingTextComplete() then
		color = crayon:GetThresholdHexColor(numQuestsDone / numQuests)
		text = text .. crayon:Colorize(color, numQuestsDone)
	end
	if self:IsShowingTextComplete() and (self:IsShowingTextCurrent() or self:IsShowingTextTotal()) then
		text = text .. crayon:Colorize(color, "/")
	end
	if self:IsShowingTextCurrent() then
		color = crayon:GetThresholdHexColor((maxQuests - numQuests) / maxQuests)
		text = text .. crayon:Colorize(color, numQuests)
	end
	if self:IsShowingTextCurrent() and self:IsShowingTextTotal() then
		text = text .. crayon:Colorize(color, "/")
	end
	if self:IsShowingTextTotal() then
		text = text .. crayon:Colorize(color, maxQuests)
	end
	if (self:IsShowingTextTotal() or self:IsShowingTextComplete() or self:IsShowingTextCurrent()) and self:IsShowingTextLastMessage() and self.lastquestmessage ~= "" then
		text = text .. " m: "
	end
	if self:IsShowingTextLastMessage() and self.lastquestmessage ~= "" then
		text = text .. self.lastquestmessage
	end
	self:SetText(text)
end

function QuestsFu:OnTooltipUpdate()
	if quixote:GetNumQuests() > 0 then
		if not self:IsShowingArea() then
			cat = tablet:AddCategory('columns', 2, 'child_func', 'QuestClick', 'child_arg1', self)
			for _,questid in quixote:QuestsByLevel() do
				self:AddQuestToCategory(cat, questid, true, self:IsShowingDescription())
			end
		else
			local currentZone = GetZoneText()
			local zonename
			local cat
			for _, zone in quixote:IterZones() do
				-- If showing area headers and either not showing only the current area's quests or this is the current area.
				if self:IsShowingArea() and ((currentZone == zone) or not self:IsShowingCurrentAreaOnly() or (self:IsShowingClassAnyway() and UnitClass("player") == zone)) then
					local r,g,b = 0.749,1,0.749
					zonename = zone
					if zone == UnitClass("player") then 
						r, g, b = babbleclass:GetColor(zone)
					else
						if self:IsShowingLevelsZone() and tourist:IsZoneOrInstance(zone) then
							local high, low = tourist:GetLevel(zone)
							if high > 0 and low > 0 then
								zonename = string.format("%s (%d-%d)", zone, high, low)
							end
						end
						if self:IsShowingDifficulty() and tourist:IsZoneOrInstance(zone) then
							r, g, b = tourist:GetLevelColor(zone)
						end
					end
					cat = tablet:AddCategory(
						'id', zone, 'text', zonename,
						'columns', 2,
						'hideBlankLine', true, 'showWithoutChildren', true,
						'checked', true, 'hasCheck', true, 'checkIcon', self.db.char.hidden[zone] and "Interface\\Buttons\\UI-PlusButton-Up" or "Interface\\Buttons\\UI-MinusButton-Up",
						'func', 'ToggleCategory', 'arg1', self, 'arg2', zone,
						'textR', r, 'textG', g, 'textB', b,
						'child_func', 'QuestClick', 'child_arg1', self
					)
					-- If we're not hiding the zone and it's either the current zone or we're not showing only the current zone.
					if (not self.db.char.hidden[zone]) and ((not self:IsShowingCurrentAreaOnly()) or (currentZone == zone) or (self:IsShowingClassAnyway() and UnitClass("player") == zone)) then
						for _,questid in quixote:QuestsByZone(zone) do
							self:AddQuestToCategory(cat, questid, false, self:IsShowingDescription())
						end
					end
				end
			end
		end
	end
	tablet:SetHint(L["TOOLTIP_HINT"])
end

function QuestsFu:MembersOnQuest(questid)
	--Return the number of party members on a quest.  Returns zero if not in a party.
	local n = 0
	local party = GetNumPartyMembers()
	if party > 0 then
		for i = 1,party do
			n = n + (IsUnitOnQuest(questid, "party"..i) and 1 or 0)
		end
	end
	return n
end

function QuestsFu:AddQuestToCategory(cat, questref, appendZone, showObjectives, hideChecks)
	local title, level, tag, objective, complete, leaderboard, zone, questid = quixote:GetQuest(questref)
	if not title or not level then return end
	
	--Are we hiding impossible quests, and is this quest impossible?
	local questImpossible = (GetDifficultyColor(level) == QuestDifficultyColor.impossible)
	if not questImpossible or (questImpossible and self:IsShowingImpossible()) then
		local thisQuest, thisQuestColor
		local questLevel, questZone = "",""
		local currentZone = GetZoneText()
		
		local questShared = QuestsFu:MembersOnQuest(questid)
		if questShared == 0 then questShared = "" end
		if appendZone then questZone = string.format(" (%s)", zone) end
		if self:IsShowingLevelsTablet() then questLevel = self:MakeTag(level, tag) end
		if self:IsShowingDifficulty() then thisQuestColor = GetDifficultyColor(level) end
		if complete == 1 then complete = L["QUEST_DONE"]
		elseif complete == -1 then complete = L["QUEST_FAILED"]
		else complete = nil end
		
		cat:AddLine(
			'text', questLevel..title..questZone..((questShared ~= "") and "|c0000ff00<"..questShared..">|r" or ''), 'wrap', self:IsWrappingQuests(),
			'text2', complete,
			'textR', (thisQuestColor and thisQuestColor.r) or 1, 'textG', (thisQuestColor and thisQuestColor.g) or 1, 'textB', (thisQuestColor and thisQuestColor.b) or 1,
			'text2R', (complete == L["QUEST_FAILED"]) and 1 or 0, 'text2G', (complete == L["QUEST_DONE"]) and 1 or 0, 'text2B', 0,
			'func', 'QuestClick', 'arg1', self, 'arg2', questid,
			'indentation', 6,
			'checked', self.db.char.watchedQuests[title],
			'hasCheck', (not hideChecks) and self.db.char.watchedQuests[title] or false
		)

		-- If we're showing descriptions/objectives and it's either the current zone or we're not showing descriptions only for the current zone.
		if showObjectives and ((currentZone == zone) or not self:IsShowingCurrentAreaDescriptionOnly()) then
			-- If we know of leaderboard objectives:
			if leaderboard > 0 then
				for i=1,leaderboard do
					local description, qtype, numGot, numNeeded, done = quixote:GetQuestStatusById(questid, i)
					if not (done and not self:IsShowingCompletedObjectives()) then
						local r,g,b
						if self:IsColoringObjectives() then
							if numGot == nil then self:Print("The author of QuestsFu would very much like to know if you see this message.  Email the rest of this line to him at kemayo at gmail dot com.", quixote:GetQuestStatusById(questid, i)) end
							r,g,b = crayon:GetThresholdColor((qtype == 'reputation' and quixote:GetReactionLevel(numGot) or numGot) / (qtype == 'reputation' and quixote:GetReactionLevel(numNeeded) or numNeeded))
						end
						local party = ''
						for i=1, GetNumPartyMembers(), 1 do
							local unit = 'party'..i
							local n = quixote:GetPartyQuestStatus(unit, title, description)
							if n and type(n) == 'number' then
								if n == numNeeded then n = 'd' end
								party = string.format("%s|c00%s%s:%s|r ", party, babbleclass:GetHexColor(UnitClass(unit)), string.sub(UnitName(unit), 1, 3), n)
							end
						end
						cat:AddLine(
							'text', '  '..description,
							'text2', party .. (done and L["QUEST_DONE"] or string.format("%s/%s", numGot, numNeeded)),
							'textR', r or ((thisQuestColor and thisQuestColor.r) or 1), 'textG', g or ((thisQuestColor and thisQuestColor.g) or 1), 'textB', b or ((thisQuestColor and thisQuestColor.b) or 1),
							'text2R', r or ((thisQuestColor and thisQuestColor.r) or 1), 'text2G', g or ((thisQuestColor and thisQuestColor.g) or 1), 'text2B', b or ((thisQuestColor and thisQuestColor.b) or 1),
							'size', tablet:GetNormalFontSize()-2, 'size2', tablet:GetNormalFontSize()-2,
							'arg2', questid,
							'indentation', 12
						)
					end
				end
			-- Otherwise, if the quest is incomplete (or complete and we're showing completed objectives), add the generic objective.
			elseif not (complete and not self:IsShowingCompletedObjectives()) then
				cat:AddLine(
					'text', objective, 'wrap', true,
					'textR', (thisQuestColor and thisQuestColor.r) or 1, 'textG', (thisQuestColor and thisQuestColor.g) or 1, 'textB', (thisQuestColor and thisQuestColor.b) or 1,
					'size', tablet:GetNormalFontSize()-2,
					'arg2', questid,
					'indentation', 12
				)
			end
		end
	end
end

function QuestsFu:ToggleCategory(id, button)
	if self.db.char.hidden[id] then
		self.db.char.hidden[id] = false
	else
		self.db.char.hidden[id] = true
	end
	-- Refresh in place
	self:UpdateTooltip()
end

--TheFly contributed the base of this, and it's still using much of the logic he provided
function QuestsFu:QuestClick(questid, button)
	if IsAltKeyDown() then
		-- Shift-click toggles quest-watch on this quest.
		if self.db.profile.useOwnTracker then
			local questName = quixote:GetQuestById(questid)
			if self.db.char.watchedQuests[questName] then
				self.db.char.watchedQuests[questName] = nil
			else
				self.db.char.watchedQuests[questName] = true
			end
			self:UpdateTracker()
		else
			if IsQuestWatched(questid) then
				RemoveQuestWatch(questid)
			else
				-- Set error if no objectives
				if GetNumQuestLeaderBoards(questid) == 0 then
					UIErrorsFrame:AddMessage(QUEST_WATCH_NO_OBJECTIVES, 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME)
					return
				end
				-- Set an error message if trying to show too many quests
				if GetNumQuestWatches() >= MAX_WATCHABLE_QUESTS then
					UIErrorsFrame:AddMessage(format(QUEST_WATCH_TOO_MANY, MAX_WATCHABLE_QUESTS), 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME)
					return
				end
				AddQuestWatch(questid)
			end
			QuestWatch_Update()
		end
	elseif IsShiftKeyDown() and IsControlKeyDown() and ChatFrameEditBox:IsVisible() then
		-- Add the quest objectives to the chat editbox, if it's open.
		_,_,_,_,_,leaderboard = quixote:GetQuestById(questid)
		local i = 1
		for i=1,leaderboard do
			local description, qtype, numGot, numNeeded = quixote:GetQuestStatusById(questid, i)
			ChatFrameEditBox:Insert(string.format("{%s %s/%s} ", description, numGot, numNeeded))
		end
	elseif IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
		-- Add quest title to the chat editbox if it's open.
		local name,level,tag = quixote:GetQuestById(questid)
		ChatFrameEditBox:Insert(self:MakeTag(level, tag)..name)
	elseif IsControlKeyDown() then
		-- Share the quest with party members.
		local wasSelected = GetQuestLogSelection()
		SelectQuestLogEntry(questid)
		if (GetQuestLogPushable() and GetNumPartyMembers() > 0) then
			QuestLogPushQuest()
		end
		SelectQuestLogEntry(questid)
	else
		self:ShowLog(questid)
	end
end

function QuestsFu:OnClick()
	ToggleQuestLog()
end

function QuestsFu:LoadWatchedQuests()
	if self.db.profile.useOwnTracker then
		self:UpdateTracker()
	else
		if not (GetNumQuestWatches() > 0) then
			for questName in pairs(self.db.char.watchedQuests) do
				local questId = quixote:GetQuestByName(questName)
				if questId then
					AddQuestWatch(questId)
				else
					self.db.char.watchedQuests[questName] = nil
				end
			end
		end
		QuestWatch_Update()
	end
	self.loadedWatchedQuests = true
end

function QuestsFu:SaveWatchedQuests()
	if self.db.profile.useOwnTracker then
		self:UpdateTracker()
	else
		if self.loadedWatchedQuests then
			self.db.char.watchedQuests = {}
			for i = 1, GetNumQuestWatches() do
				local questName = quixote:GetQuestById(GetQuestIndexForWatch(i))
				if questName then self.db.char.watchedQuests[questName] = true end
			end
		end
	end
end

function QuestsFu:AddQuestWatch(questid)
	if self.db.profile.useOwnTracker then
		local questName = quixote:GetQuestById(questid)
		self.db.char.watchedQuests[questName] = true
	end
	self:SaveWatchedQuests()
	QuestWatchFrame:Hide()
end

function QuestsFu:RemoveQuestWatch(questid)
	if self.db.profile.useOwnTracker then
		local questName = quixote:GetQuestById(questid)
		self.db.char.watchedQuests[questName] = nil
	end
	self:SaveWatchedQuests()
end

function QuestsFu:Quest_Lost(name, id)
	if self.db.profile.useOwnTracker then
		self.db.char.watchedQuests[name] = nil
	end
end

function QuestsFu:UpdateTracker()
	if self.db.profile.useOwnTracker then
		if not tablet:IsRegistered('QuestsFu_Tracker') then
			tablet:Register('QuestsFu_Tracker', 'detachedData', self.db.profile.tracker_data, 'cantAttach', true, 'dontHook', true, 'hideWhenEmpty', true,
				'children', function()
					tablet:SetTitle("Tracker")
					local cat = tablet:AddCategory('columns', 2, 'child_func', 'QuestClick', 'child_arg1', self)
					for questName in pairs(self.db.char.watchedQuests) do self:AddQuestToCategory(cat, questName, false, true, true) end
				end)
			tablet:Open('QuestsFu_Tracker')
		end
		tablet:Refresh('QuestsFu_Tracker')
		QuestWatchFrame:Hide()
	end
end

function QuestsFu:ShowLog(questid)
	if QuestLogFrame:IsVisible() then
		if self.lastIndex == questid then
			HideUIPanel(QuestLogFrame)
		end
	else
		ShowUIPanel(QuestLogFrame)
	end
	local numEntries = GetNumQuestLogEntries()
	if (numEntries > QUESTS_DISPLAYED) then
		if (questid < numEntries - QUESTS_DISPLAYED) then
			FauxScrollFrame_SetOffset(QuestLogListScrollFrame, questid - 1)
			QuestLogListScrollFrameScrollBar:SetValue((questid - 1) * QUESTLOG_QUEST_HEIGHT)
		else
			FauxScrollFrame_SetOffset(QuestLogListScrollFrame, numEntries - QUESTS_DISPLAYED)
			QuestLogListScrollFrameScrollBar:SetValue((numEntries - QUESTS_DISPLAYED) * QUESTLOG_QUEST_HEIGHT)
		end
	end

	SelectQuestLogEntry(questid)
	QuestLog_SetSelection(questid)
	self.lastIndex = questid
end
